from winning.lattice_calibration import state_price_implied_ability, dividend_implied_ability
from winning.lattice import skew_normal_density, center_density
from winning.lattice_conventions import STD_UNIT, STD_SCALE, STD_L, STD_A


# Specializing the ratings race problem to N(a,1) horses


def std_state_price_implied_ability(prices):
    """ Ability implied from state prices

          returns:  [ float ]   List of relative abilities

    """
    # (For continuous distributions state prices are synonymous with winning probabilities)
    density = std_density(loc=0)
    return state_price_implied_ability(prices=prices, density=density)

def std_dividend_implied_ability(dividends):
    """ Implied ability using skewed normal performances, from inverse win probabilities """
    return dividend_implied_ability(dividends=dividends, density=std_density(loc=0.0))

def std_density(loc=0):
    density = skew_normal_density(L=STD_L, unit=STD_UNIT, loc=loc, scale=STD_SCALE, a=STD_A)
    return center_density(density)






from winning.lattice import symmetric_lattice

def densitiesPlot( densities, unit, legend = None ):
    import matplotlib.pyplot as plt
    L = int(len( densities[0]-1 )/2)
    pts = symmetric_lattice( L=L, unit=unit )
    for density in densities:
        plt.plot( pts, density )
    if legend is not None:
        plt.legend( legend )
    plt.grid()
    plt.show()






from scipy.stats import norm
import numpy as np
import math

#########################################################################################
#   Operations on univariate atomic distributions supported on evenly spaced points     #
#########################################################################################

def integer_shift( cdf, k ):
    """ Shift cdf to the *right* so it represents the cdf for Y ~ X + k*unit
    :param cdf:
    :param k:     int    Number of lattice points
    :return:
    """
    if k < 0:
        return np.append( cdf[ abs(k):], cdf[-1]*np.ones( abs(k) ) )
    elif k==0:
        return cdf
    else:
        return np.append( np.zeros(k), cdf[:-k] )

def fractional_shift(cdf, x):
    """ Shift cdf to the *right* so it represents the cdf for Y ~ X + x*unit
    :param cdf:
    :param x:     float    Number of lattice points to shift (need not be integer)
    :return:
    """
    ( l, lc), ( u, uc) = _low_high( x )
    return lc * integer_shift(cdf, l) + uc * integer_shift(cdf, u)

def _low_high( offset ):
    """ Represent a float offset as combination of two discrete ones """
    l = math.floor( offset )
    u = math.ceil( offset )
    r = offset - l
    return (l, 1-r ), (u, r)

def fractional_shift_density( density, x ):
    """ Shift pdf to the *right* so it represents the pdf for Y ~ X + x*unit """
    cdf                = pdf_to_cdf(density)
    shifted_cdf        = fractional_shift( cdf, x )
    return cdf_to_pdf(shifted_cdf)

def center_density( density ):
    """ Shift density to near its mean """
    m = mean_of_density(density, unit=1.0)
    return fractional_shift_density( density, -m )

def mean_of_density( density, unit ):
    L   = int((len( density)-1) /2)
    pts = symmetric_lattice( L=L, unit=unit )
    return np.inner( density, pts )

def symmetric_lattice(L, unit):
    assert isinstance(L,int),"Expecting L to be integer"
    return unit*np.linspace( -L, L, 2*L+1 )

#############################################
#   Simple family of skewed distributions   #
#############################################

def skew_normal_density( L, unit, loc=0, scale=1.0, a=2.0):
    """ Skew normal as a lattice density """
    lattice = symmetric_lattice(L=L, unit=unit)
    density = np.array( [ _unnormalized_skew_cdf(x, loc=loc, scale=scale, a=a ) for x in lattice])
    density = density / np.sum( density )
    return density

def _unnormalized_skew_cdf(x, loc=0, scale=1.0, a=2.0):
    """ Proportional to skew-normal density
    :param x:
    :param loc:    location
    :param scale:  scale
    :param a:      controls skew (a>0 means fat tail on right)
    :return:  np.array length 2*L+1
    """
    t = (x-loc) / scale
    return 2 / scale * norm.pdf(t) * norm.cdf(a*t)



# -------  Nothing below here depends on the unit chosen -------------



#############################################
#  Order statistics on lattices             #
#############################################

def pdf_to_cdf( density ):
    """  Prob( X <= k*unit )    """
    return np.cumsum( density )

def cdf_to_pdf(cumulative):
    """ Given cumulative distribution on lattice, return the pdf """
    prepended = np.insert( cumulative, 0, 0.)
    return np.diff( prepended )

def winner_of_many( densities, multiplicities = None):
    """ The PDF of the minimum of the random variables represented by densities
    :param   densities:  [ np.array   ]
    :return: np.array
    """
    d  = densities[0]
    multiplicities = multiplicities or [ None for _ in densities ]
    m = multiplicities[0]
    for d2, m2 in zip( densities[1:], multiplicities[1:] ):
        d, m = _winner_of_two_pdf( d, d2, multiplicityA=m, multiplicityB = m2 )
    return d, m

def sample_from_cdf( cdf, nSamples ):
    """ Monte Carlo sample """
    rvs = np.random.rand( nSamples )
    return [ sum( [ rv>c for c in cdf ] ) for rv in rvs ]

def sample_winner_of_many( densities, nSamples = 5000 ):
    """ The PDF of the minimum of the integer random variables represented by densities, by Monte Carlo """
    cdfs = [pdf_to_cdf(density) for density in densities]
    cols = [sample_from_cdf(cdf, nSamples) for cdf in cdfs]
    rows = map( list, zip( *cols ))
    D    = [ min( row ) for row in rows ]
    density = np.bincount( D, minlength=len( densities[0] ) ) / (1.0*nSamples)
    return density

def expected_payoff( density, densityAll, multiplicityAll, cdf = None, cdfAll = None):
    """ Returns expected _conditional_payoff_against_rest broken down by score,
       where _conditional_payoff_against_rest is 1 if we are better than rest (lower) and 1/(1+multiplicity) if we are equal

    """
    # Use np.sum( expected_payoff ) for the expectation
    if cdf is None:
        cdf      = pdf_to_cdf(density)
    if cdfAll is None:
        cdfAll   = pdf_to_cdf(densityAll)
    if density is None:
        density = cdf_to_pdf(cdf)
    if densityAll is None:
        densityAll = cdf_to_pdf(cdfAll)
    S       = 1 - cdfAll
    S1      = 1 - cdf
    Srest   = ( S + 1e-18 ) / ( S1 + 1e-6 )
    cdfRest = 1 - Srest

    # Multiplicity inversion (uses notation from blog post)
    # This is written up in my blog post
    m       = multiplicityAll
    f1      = density
    m1      = 1.0
    fRest   = cdf_to_pdf(cdfRest)
                                              # numer = m*f1*Srest + m*(f1+S1)*fRest - m1*f1*( Srest + fRest )
                                              # denom = fRest*(f1+S1)
    numer   = m*f1*Srest + m*(f1+S1)*fRest - m1*f1*( Srest + fRest )
    denom   = fRest*(f1+S1)
    multiplicityLeftTail = (1e-18 + numer ) / ( 1e-18 + denom )
    multiplicityRest = multiplicityLeftTail

    T1      = (S1 +1.0e-18) / (f1 + 1e-6 )      # This calculation is more stable on the right tail. It should tend to zero eventually
    Trest   = (Srest + 1e-18) / ( fRest + 1e-6 )
    multiplicityRightTail = m*Trest / (1 + T1) + m - m1 * (1 + Trest ) / (1 + T1 )

    k       = list( f1 == max(f1) ).index( True )
    multiplicityRest[k:] = multiplicityRightTail[k:]

    return _conditional_payoff_against_rest(density = density, densityRest = None, multiplicityRest = multiplicityRest, cdf = cdf, cdfRest = cdfRest)

def _winner_of_two_pdf( densityA, densityB, multiplicityA = None, multiplicityB = None, cdfA = None, cdfB = None):
    """ The PDF of the minimum of two random variables represented by densities
    :param   densityA:   np.array
    :param   densityB:   np.array
    :return: density, multiplicity
    """
    cdfA    = pdf_to_cdf(densityA)
    cdfB    = pdf_to_cdf(densityB)
    cdfMin  = 1 - np.multiply( 1 - cdfA, 1- cdfB )
    density = cdf_to_pdf(cdfMin)
    L       = int((len( density )-1) / 2)
    if multiplicityA is None:
        multiplicityA = np.ones( 2*L+1 )
    if multiplicityB is None:
        multiplicityB = np.ones( 2*L+1 )

    winA, draw, winB = _conditional_win_draw_loss( densityA, densityB, cdfA, cdfB )
    multiplicity  = ( winA*multiplicityA + draw*(multiplicityA+multiplicityB) + winB*multiplicityB +1e-18) / ( winA+draw+winB+1e-18)
    return density, multiplicity

def _conditional_win_draw_loss(densityA, densityB, cdfA, cdfB):
    """ Conditional win, draw and loss probability lattices for a two ratings race """
    win   = densityA * ( 1 - cdfB )
    draw  = densityA * densityB
    lose  = densityB * ( 1 - cdfA )
    return win, draw, lose

def _conditional_payoff_against_rest(density, densityRest, multiplicityRest, cdf = None, cdfRest = None):
    """ Returns expected _conditional_payoff_against_rest broken down by score, where _conditional_payoff_against_rest is 1 if we are better than rest (lower) and 1/(1+multiplicity) if we are equal """
    # use np.sum( _conditional_payoff_against_rest) for the expectation
    if cdf is None:
        cdf = pdf_to_cdf(density)
    if cdfRest is None:
        cdfRest = pdf_to_cdf(densityRest)
    if density is None:
        density = cdf_to_pdf(cdf)
    if densityRest is None:
        densityRest = cdf_to_pdf(cdfRest)

    win, draw, loss = _conditional_win_draw_loss(density, densityRest, cdf, cdfRest)
    return win + draw / (1+multiplicityRest )

def densities_and_coefs_from_offsets( density, offsets ):
    """ Given a density and a list of offsets (which might be non-integer) this
        returns a list of translated densities

    :param density:  np.ndarray
    :param offsets: [ float ]
    :return: [ np.ndarray ]
    """
    cdf             = pdf_to_cdf(density)
    coefs           = [ _low_high( offset ) for offset in offsets ]
    cdfs            = [lc * integer_shift(cdf, l) + uc * integer_shift(cdf, u) for (l, lc), (u, uc) in coefs]
    expected_len    = len(density)
    pdfs = [ cdf_to_pdf(cdf) for cdf in cdfs]
    return pdfs, coefs


def densities_from_offsets( density, offsets ):
    return densities_and_coefs_from_offsets( density, offsets)[0]

def state_prices_from_offsets( density, offsets ):
    """ Returns a list of state prices for a race where all horses have the same density
        up to a translation (the offsets)
    """
    # See the paper for a definition of state price
    # Be aware that this may fail if offsets provided are integers rather than float
    densities = densities_from_offsets( density, offsets )
    densityAll, multiplicityAll = winner_of_many( densities )
    return implicit_state_prices( density, densityAll = densityAll, multiplicityAll=multiplicityAll, offsets=offsets )

def implicit_state_prices( density, densityAll, multiplicityAll = None, cdf = None, cdfAll = None, offsets = None ):
    """ Returns the expected _conditional_payoff_against_rest as a function of location changes in cdf """

    L = int((len( density )-1)/2)
    if cdf is None:
        cdf = pdf_to_cdf( density )
    if cdfAll is None:
        cdfAll = pdf_to_cdf(densityAll)
    if multiplicityAll is None:
        multiplicityAll = np.ones( 2*L+1 )
    if offsets is None:
        offsets = range( int(-L/2), int(L/2) )
    implicit = list()
    for k in offsets:
        if k==int( k ):
            offset_cdf = integer_shift( cdf, k )
            ip         = expected_payoff( density = None, densityAll = densityAll, multiplicityAll=multiplicityAll, cdf = offset_cdf, cdfAll = cdfAll)
            implicit.append( np.sum( ip ) )
        else:
            (l, l_coef ), ( r, r_coef) = _low_high( k )
            offset_cdf_left            = integer_shift(cdf, l)
            offset_cdf_right           = integer_shift(cdf, r)
            ip_left   = expected_payoff(density = None, densityAll = densityAll, multiplicityAll=multiplicityAll, cdf = offset_cdf_left, cdfAll = cdfAll)
            ip_right  = expected_payoff(density = None, densityAll = densityAll, multiplicityAll=multiplicityAll, cdf = offset_cdf_right, cdfAll = cdfAll)
            implicit.append( l_coef*np.sum( ip_left ) + r_coef*np.sum( ip_right ) )

    return implicit







from winning.lattice_calibration import state_price_implied_ability, dividend_implied_ability, ability_implied_dividends
from winning.lattice import center_density, skew_normal_density
from winning.lattice_conventions import ALT_UNIT, ALT_L, ALT_SCALE, ALT_A

# Specializing the ratings race problem to skew-normally distributed horses


def skew_state_price_implied_ability(prices: [float]) -> [float]:
    """ Calibrate to a skewed performance distribution

            prices:    state prices (very similar to probabilities of winning)
            returns:   implied ratings corresponding to location parameters (i.e. relative translations of perf distribution)

    """
    density = skew_normal_density(L=ALT_L, unit=ALT_UNIT, loc=0., scale=ALT_SCALE, a=ALT_A)
    return state_price_implied_ability(prices=prices, density=density)

def skew_density(loc):
    """ An canonical choice of skew-normal density using hard-coded parameters """
    density = skew_normal_density(L=ALT_L, unit=ALT_UNIT, loc=loc, scale=ALT_SCALE, a=ALT_A)
    return center_density( density )

def skew_dividend_implied_ability(dividends):
    """ Implied ability using skewed normal performances, from inverse win probabilities """
    return dividend_implied_ability(dividends=dividends, density=skew_density(loc=0.0))

def skew_ability_implied_dividends(ability):
    """ Inverse probabilities based on a canonical choice of skew normal density """
    return ability_implied_dividends(ability, density=skew_density(loc=0.0))





from winning.lattice import state_prices_from_offsets, densities_and_coefs_from_offsets,\
    winner_of_many, expected_payoff, densities_from_offsets, implicit_state_prices
import pandas as pd
import numpy as np


#################################################################
#                                                               #
#      Implements the "method of multiplicity inversion" that   #
#      quickly solves the ratings race problem                    #
#                                                               #
#################################################################


def make_nan_2000(x):
    """ Longshots """
    if pd.isnull(x):
        return 2000.
    else:
        return x


def normalize(p):
    """ Naive renormalization of probabilities """
    S = sum(p)
    return [pr / S for pr in p]

def prices_from_dividends(dividends):
    """ Risk neutral probabilities using naive renormalization """
    return normalize([1. / make_nan_2000(x) for x in dividends])


def dividends_from_prices(prices):
    """ Australian style dividends """
    return [1. / d for d in normalize(prices)]

def normalize_dividends(dividends):
    return dividends_from_prices(prices_from_dividends(dividends))


def dividend_implied_ability(dividends, density):
    """ Infer risk-neutral implied_ability from Australian style dividends

    :param dividends:    [ 7.6, 12.0, ... ]
    :return: [ float ]   Implied ability

    """
    p = prices_from_dividends(dividends)
    return state_price_implied_ability(prices=p, density=density)

def state_price_implied_ability(prices, density):
    """ Calibrate offsets (translations of the performance density) to match state prices """
    implied_offsets_guess = [0 for _ in prices]
    L = int((len(density) - 1) / 2)
    offset_samples = list(range(int(-L / 2), int(L / 2)))[::-1]
    ability = implied_ability(prices=prices, density=density, \
                              offset_samples=offset_samples, implied_offsets_guess=implied_offsets_guess, nIter=3)
    return ability


def ability_implied_dividends(ability, density):
    """ Return inverse state prices
    :param ability:   [ float ]
    :param density:  [ float ]
    :return: [ 7.6, 12.3, ... ]
    """
    state_prices = state_prices_from_offsets(density=density, offsets=ability)
    return [1. / sp for sp in state_prices]


def implied_ability(prices, density, offset_samples=None, implied_offsets_guess=None, nIter=3, verbose=False,
                    visualize=False):
    """
    This is the main routine.
    It finds location translations of a fixed performance densitym so as to replicate given state prices for winning
    See the paper for details.

        offset_samples   Optionally supply a list of offsets which are used in the interpolation table  a_i -> p_i

    """

    L = int((len(density) - 1) / 2)
    if offset_samples is None:
        offset_samples = list(range(int(-L / 2), int(L / 2)))[
                         ::-1]
    else:
        _assert_descending(offset_samples)

    if implied_offsets_guess is None:
        implied_offsets_guess = list(range(int(L/3)))

    # First guess at densities
    densities, coefs = densities_and_coefs_from_offsets(density, implied_offsets_guess)
    densityAllGuess, multiplicityAllGuess = winner_of_many(densities)
    densityAll = densityAllGuess.copy()
    multiplicityAll = multiplicityAllGuess.copy()
    guess_prices = [np.sum(expected_payoff(density, densityAll, multiplicityAll, cdf=None, cdfAll=None)) for density in
                    densities]

    for _ in range(nIter):
        if visualize:
            from winning.lattice_plot import densitiesPlot
            # temporary hack to check progress of optimization
            densitiesPlot([densityAll] + densities, unit=0.1)
        implied_prices = implicit_state_prices(density=density, densityAll=densityAll, multiplicityAll=multiplicityAll,
                                               offsets=offset_samples)
        implied_offsets = np.interp(prices, implied_prices, offset_samples)
        densities = densities_from_offsets(density, implied_offsets)
        densityAll, multiplicityAll = winner_of_many(densities)
        guess_prices = [np.sum(expected_payoff(density, densityAll, multiplicityAll, cdf=None, cdfAll=None)) for density
                        in densities]
        approx_prices = [np.round(pri, 3) for pri in prices]
        approx_guesses = [np.round(pri, 3) for pri in guess_prices]
        if verbose:
            print(zip(approx_prices, approx_guesses)[:5])

    return implied_offsets


def _assert_descending(xs):
    for d in np.diff(xs):
        if d > 0:
            raise ValueError("Not descending")
